package zdb.client;

import fr.zenexity.json.JSON;
import java.lang.reflect.AccessibleObject;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.UUID;

import net.sf.json.JSONObject;
import net.sf.json.JSONSerializer;

import org.apache.lucene.analysis.Analyzer;
import org.apache.lucene.analysis.standard.StandardAnalyzer;
import org.apache.lucene.document.DateTools;
import org.apache.lucene.index.Term;
import org.apache.lucene.queryParser.ParseException;
import org.apache.lucene.queryParser.QueryParser;
import org.apache.lucene.search.BooleanQuery;
import org.apache.lucene.search.MatchAllDocsQuery;
import org.apache.lucene.search.Query;
import org.apache.lucene.search.Sort;
import org.apache.lucene.search.TermQuery;
import org.apache.lucene.search.BooleanClause.Occur;

import zdb.client.annotations.ID;
import zdb.client.annotations.Index;
import zdb.exceptions.UnexpectedException;
import zdb.exceptions.ZDBException;

public abstract class Bucket {

    public static abstract class DocumentQuery {

        Bucket bucket;
        String[] orderBy = new String[0];
        Query query;

        public DocumentQuery(Bucket bucket, Query query) {
            this.query = query;
            this.bucket = bucket;
        }

        public abstract List<Document> all(String retrieve);

        public abstract List<Document> page(int from, int size, String retrieve);

        public abstract Document one(String retrieve);

        public abstract int count();

        public DocumentQuery orderBy(String... fields) {
            this.orderBy = fields;
            return this;
        }

        protected Sort getSort() {
            return new Sort(orderBy);
        }
    }

    public static class ObjectQuery<T> {

        DocumentQuery documentQuery;
        Class<T> clazz;

        public ObjectQuery(Bucket bucket, Query query, Class<T> clazz) {
            BooleanQuery rq = new BooleanQuery();
            rq.add(query, Occur.MUST);
            rq.add(new TermQuery(new Term("_type", Bucket.getType(clazz))), Occur.MUST);
            this.documentQuery = bucket.search(rq);
            this.clazz = clazz;
        }

        public List<T> all(String retrieve) {
            return page(0, Integer.MAX_VALUE, retrieve);
        }

        public List<T> page(int from, int size, String retrieve) {
            List<T> result = new ArrayList<T>();
            for (Document document : documentQuery.page(from, size, retrieve)) {
                result.add(documentQuery.bucket.toObject(document, clazz));
            }
            return result;
        }

        public T one(String retrieve) {
            Document document = documentQuery.one(retrieve);
            if (document == null) {
                return null;
            }
            return documentQuery.bucket.toObject(document, clazz);
        }

        public int count() {
            return documentQuery.count();
        }

        public ObjectQuery<T> orderBy(String... fields) {
            documentQuery = documentQuery.orderBy(fields);
            return this;
        }
    }

    public abstract boolean exists(String id);

    public abstract Document get(String id, boolean indexes);

    public abstract void put(Document document);

    public abstract void delete(String id);

    public abstract DocumentQuery search(Query query);

    public DocumentQuery search(String query) throws ParseException {
        Query q = new QueryParser("", new StandardAnalyzer()).parse(query);
        return search(q);
    }

    public DocumentQuery search() {
        return search(new MatchAllDocsQuery());
    }

    public <T> T get(String id, Class<T> type) {
        Document document = get(id, false);
        if (document == null) {
            return null;
        }
        return toObject(document, type);
    }

    public <T> void put(Object object) {
        Document document = toDocument(object);
        put(document);
    }

    public void delete(Object object) {
        Document document = toDocument(object);
        delete(document.id);
    }

    public <T> ObjectQuery<T> search(Query query, Class<T> type) {
        return new ObjectQuery<T>(this, query, type);
    }

    public <T> ObjectQuery<T> search(String query, Class<T> type) throws ParseException {
        return search( query, type, new StandardAnalyzer());
    }

    public <T> ObjectQuery<T> search(String query, Class<T> type, Analyzer analyzer ) throws ParseException {
        Query q = new QueryParser("", analyzer).parse(query);
        return new ObjectQuery<T>(this, q, type);
    }

    public <T> ObjectQuery<T> search(Class<T> type) {
        return new ObjectQuery<T>(this, new MatchAllDocsQuery(), type);
    }
    /* ~~~~~~~~~~~~~~~~~~~~ Automatic java binder */
    // caches to speed up object serialization/deserialization
    private static Map<Class, String> types = new HashMap<Class, String>();
    private static Map<Class, Field> ids = new HashMap<Class, Field>();
    private static Map<Class, Map<AccessibleObject, Index>> indexes = new HashMap<Class, Map<AccessibleObject, Index>>();
    /** cache classes that provide their own Json unmarshalling method */
    private static Map<Class, Method> fromJSONs = new HashMap<Class, Method>();
    /** cache classes that provide their own Json marshalling method */
    private static Map<Class, Method> toJSONs = new HashMap<Class, Method>();

    static <T> String getType(Class<T> clazz) {
        if (!types.containsKey(clazz)) {
            zdb.client.annotations.Document doc = clazz.getAnnotation(zdb.client.annotations.Document.class);
            if (doc == null)
            	throw new ZDBException ("The class "+clazz.getName()+" is not annotated with @Document");
            String type = doc.type();
            if (type.equals("_")) {
                type = clazz.getSimpleName();
            }
            types.put(clazz, type);
        }
        return types.get(clazz);
    }

    static <T> Field getID(Class<T> clazz) {
        if (!ids.containsKey(clazz)) {
            for (Field field : clazz.getDeclaredFields()) {
                if (field.isAnnotationPresent(ID.class) && field.getType().equals(String.class)) {
                    field.setAccessible(true);
                    ids.put(clazz, field);
                    break;
                }
            }
        }
        return ids.get(clazz);
    }

    static <T> Method getFromJSON(Class<T> clazz) {
        if (!fromJSONs.containsKey(clazz)) {
            for (Method method : clazz.getDeclaredMethods()) {
                if (method.getName().equals("fromJSON") && Modifier.isStatic(method.getModifiers()) && method.getParameterTypes().length == 1 && method.getParameterTypes()[0].equals(String.class)) {
                    method.setAccessible(true);
                    fromJSONs.put(clazz, method);
                    break;
                }
            }
        }
        return fromJSONs.get(clazz);
    }

    static <T> Method getToJSON(Class<T> clazz) {
        if (!toJSONs.containsKey(clazz)) {
            for (Method method : clazz.getDeclaredMethods()) {
                if (method.getName().equals("toJSON") && method.getParameterTypes().length == 0 && method.getReturnType().equals(String.class)) {
                    method.setAccessible(true);
                    toJSONs.put(clazz, method);
                    break;
                }
            }
        }
        return toJSONs.get(clazz);
    }

    static <T> Map<AccessibleObject, Index> getIndexes(Class<T> clazz) {
        if (!indexes.containsKey(clazz)) {
            Map<AccessibleObject, Index> result = new HashMap<AccessibleObject, Index>();
            for (Method method : clazz.getDeclaredMethods()) {
                if (method.isAnnotationPresent(Index.class)) {
                    method.setAccessible(true);
                    result.put(method, method.getAnnotation(Index.class));
                }
            }
            for (Field field : clazz.getDeclaredFields()) {
                if (field.isAnnotationPresent(Index.class)) {
                    field.setAccessible(true);
                    result.put(field, field.getAnnotation(Index.class));
                }
            }
            indexes.put(clazz, result);
        }
        return indexes.get(clazz);
    }

    public Document toDocument(Object object) {

        try {
            // New document
            Document document = new Document();
            // Set mimeType to JSON
            document.mimeType = "text/json";
            // Retrieve docType
            String type = getType(object.getClass());
            // Retrieve id and generate UUID if null
            Field idF = getID(object.getClass());
            assert idF != null;
            String id = (String) idF.get(object);
            if (id == null) {
                id = UUID.randomUUID().toString();
                idF.set(object, id);
            }
            document.id = id;
            // Transform to JSON
            String json = null;
            Method toJSON = getToJSON(object.getClass());
            if (toJSON == null) {
                json = JSON.toJSON(object);
            } else {
                json = (String) toJSON.invoke(object);
            }
            document.setContent(json);
            // Add indexes
            document.addIndex("_type", true, true, false, true, type);
            Map<AccessibleObject, Index> lindexes = getIndexes(object.getClass());
            for (AccessibleObject accessibleObject : lindexes.keySet()) {
                Object value = null;
                if (accessibleObject instanceof Field) {
                    value = ((Field) accessibleObject).get(object);
                }
                if (accessibleObject instanceof Method) {
                    value = ((Method) accessibleObject).invoke(object);
                }
                String stringValue = "";
                Index index = lindexes.get(accessibleObject);
                String name = index.name();
                if (name.equals("_")) {
                    if (accessibleObject instanceof Field) {
                        name = ((Field) accessibleObject).getName();
                    }
                    if (accessibleObject instanceof Method) {
                        name = ((Method) accessibleObject).getName();
                        if (name.startsWith("get") && name.length() > 3) {
                            name = name.substring(3);
                            name = name.substring(0, 1).toLowerCase() + name.substring(1);
                        }
                    }
                }
                if (value != null) {
                    // add document indices for collections and array also
                    addIndexes(name, index, document, value);
                }
            }
            return document;
        } catch (IllegalAccessException e) {
            throw new UnexpectedException("OOps", e);
        } catch (InvocationTargetException e) {
            throw new UnexpectedException("OOps", e);
        }
    }

    /**
     * internal - create a lucene index "name" with the given value
     * value maybe a single-valued type ( String, Date) or a container: array or collection.
     * for a container, each value is indexed individually
     * @param name String the index identifier
     * @param index Index information
     * @param document the document 
     * @param value the value to index
     */
    public void addIndexes(String name, Index index, Document document, Object value) {

        if (value.getClass().isArray()) {
            Object[] values = (Object[]) value;
            for (int i = 0; i < values.length; i++) {
                addIndexes(name, index, document, values[i]);
            }
        } else if (Collection.class.isAssignableFrom(value.getClass())) {
            Iterator ite = ((Collection) value).iterator();
            while (ite.hasNext()) {
                addIndexes(name, index, document, ite.next());
            }
        } else if (value instanceof Date) {
            String valueToIndex = DateTools.dateToString((Date) value, DateTools.Resolution.MILLISECOND);
            document.addIndex( new Document.Index( name, valueToIndex, index));
        } else {
            document.addIndex( new Document.Index( name, value.toString(), index));
        }
    }

    public <T> T toObject(Document document, Class<T> type) {
        try {
            T bean = null;
            // todo : methods caching should be handled by JSON lib
            Method fromJSON = getFromJSON(type);
            String json = document.content();
            if (fromJSON == null) {
                bean = (T) JSON.fromJSON(json, type);
            } else {
                bean = (T) fromJSON.invoke(null, json);
            }
            Field id = getID(type);
            id.set(bean, document.id);
            document.release();
            return bean;
        } catch (Exception e) {
            throw new UnexpectedException("Cannot read object", e);
        }
    }
}
