package zdb.client;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import org.apache.lucene.search.Query;

import zdb.core.Store;
import zdb.core.Store.QueryResult;
import zdb.exceptions.UnexpectedException;

public class DirectBucket extends Bucket {

    public String name;

    public DirectBucket(String name) {
        this.name = name;
        Store.createBucket(name);
    }

    @Override
    public Document get(String id, boolean indexes) {
        final Store.ReadResult readResult = Store.read(name, id, indexes);
        final Document document = new Document();
        document.id = readResult.id;
        document.mimeType = readResult.mimeType;
        document.size = readResult.size;
        document.indexes = readResult.indexes;
        document.documentObserver = new Document.DocumentObserver() {

            public void onClose() {
                try {
                    if (document.readIs != null) {
                        document.readIs.close();
                    }
                } catch (IOException e) {
                    throw new UnexpectedException("Can't close", e);
                }
                readResult.release();
            }

            public void onRead() {
                if (document.readIs != null) {
                    return;
                }
                try {
                    document.readIs = new FileInputStream(readResult.file);
                } catch (FileNotFoundException e) {
                    throw new UnexpectedException("Cant read file", e);
                }
            }
        };
        return document;
    }

    @Override
    public void put(Document document) {
        if (document.id == null) {
            throw new RuntimeException("ID must be set");
        }
        if (document.mimeType == null) {
            throw new RuntimeException("mimeType must be set");
        }
        if (document.writeIs == null) {
            throw new RuntimeException("You must set writeIs when performing put operations");
        }
        Store.save(name, document.id, document.writeIs, document.mimeType, document.serializeIndexes(), null);
    }

    @Override
    public void delete(String id) {
        Store.delete(name, id);
    }

    @Override
    public DocumentQuery search(Query query) {
        return new DirectDocumentQuery(this, query);
    }

    public static class DirectDocumentQuery extends DocumentQuery {

        public DirectDocumentQuery(DirectBucket bucket, Query query) {
            super(bucket, query);
        }

        @Override
        public List<Document> all(String retrieve) {
            return page(0, Integer.MAX_VALUE, retrieve);
        }

        @Override
        public List<Document> page(int from, int size, String retrieve) {
            List<Document> result = new ArrayList<Document>();
            List<QueryResult> search = Store.search(((DirectBucket) bucket).name, query, getSort(), from, size, retrieve);
            for (QueryResult queryResult : search) {
                Document doc = ((DirectBucket) bucket).get(queryResult.docID(), true);
                if (doc != null) {
                    doc.score = queryResult.score;
                }
                result.add(doc);
            }
            return result;
        }

        @Override
        public Document one(String retrieve) {
            List<Document> result = page(0, 1, retrieve);
            if (result.size() == 0) {
                return null;
            }
            return result.get(0);
        }

        @Override
        public int count() {
            return Store.count(((DirectBucket) bucket).name, query);
        }
    }

    public boolean exists(String id) {
        return Store.exists(name, id);
    }
}
