package zdb.client;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.UnsupportedEncodingException;
import java.util.ArrayList;
import java.util.List;

import zdb.exceptions.UnexpectedException;

public class Document {

    /**
     * Document unique identifier - used by lucene index also
     */
    public String id;
    public String mimeType;
    public List<Index> indexes = new ArrayList<Index>();
    InputStream writeIs;
    InputStream readIs;
    DocumentObserver documentObserver;
    public long size;
    /**
     * score of this document for the query (if Document is part of a query results)
     */
    public float score;

   
    protected Document() {
    }

    public Document(String id, String mimeType, String content) {
        this.id = id;
        this.mimeType = mimeType;
        setContent(content);
    }
    
    public Document (String id, String mimeType, InputStream content) {
    	this.id=id;
    	this.mimeType=mimeType;
    	this.writeIs=content;
    }
 
    public void setContent(String content) {
        try {
            this.writeIs = new ByteArrayInputStream(content.getBytes("utf-8"));
        } catch (UnsupportedEncodingException ex) {
            throw new UnexpectedException("Invalid encoding", ex);
        }
    }

    public Index getIndex (String name) {
    	for (Index index : indexes) {
			if (index.name.equals(name))
				return index;
		}
    	return null;
    }
    
    public void release() {
        if (writeIs != null) {
            try {
                writeIs.close();
                writeIs = null;
            } catch (IOException ex) {
                ex.printStackTrace();
            }
        }
        if(documentObserver != null) {
        	documentObserver.onClose();
        	documentObserver = null;
        }
    }

    public InputStream rawContent () {
    	if (documentObserver!=null) documentObserver.onRead();
    	return readIs;
    }
    
    public String content() throws IOException {
        try {
        	if (documentObserver!=null) documentObserver.onRead();
        	byte[] data = new byte[(int)size];
        	readIs.read(data, 0, (int)size);
            return new String(data, "utf-8");
        } finally {
            release();
        }
    }

    /**
     * register an index for the Document
     * @param name
     * @param stored
     * @param indexed
     * @param fulltext
     * @param value
     */
    public void addIndex(String name, boolean stored, boolean indexed, boolean fulltext, boolean exact, String value) {
        Index index = new Index(name, value, stored, indexed, fulltext, exact);
        indexes.add(index);
    }

    public void addIndex( Document.Index docIndex) {
        indexes.add(docIndex);
    }
    
    protected List<String> serializeIndexes() {
        List<String> result = new ArrayList<String>();
        for (Index index : indexes) {
            StringBuilder stringBuilder = new StringBuilder();
            if (index.indexed) {
                stringBuilder.append("INDEXED;");
            }
            if (index.stored) {
                stringBuilder.append("STORED;");
            }
            if (index.fulltext) {
                stringBuilder.append("FULLTEXT;");
            }
            if (index.exact) {
                stringBuilder.append("EXACT;");
            }
            stringBuilder.append(index.name);
            stringBuilder.append("=");
            stringBuilder.append(index.value);
            result.add(stringBuilder.toString());
        }
        return result;
    }

    public static class Index {

        public String name;
        public String value;
        public boolean stored;
        public boolean indexed;
        public boolean exact;
        public boolean fulltext;

        public Index (String name, String value) {
            this.name = name;
            this.value = value.replace("\n", " ");       	
        }
        
        public Index(String name, String value, boolean stored, boolean indexed, boolean fulltext, boolean exact) {
            this.name = name;
            this.value = value.replace("\n", " ");
            this.stored = stored;
            this.indexed = indexed;
            this.fulltext = fulltext;
            this.exact = exact;
        }

        public Index( String name, String value, zdb.client.annotations.Index options ) {
            this.name = name;
            this.value = value.replace("\n", " ");
            this.indexed = options.indexed();
            this.stored = options.stored();
            this.fulltext = options.fulltext();
            this.exact = options.exact();
        }
    }
    
    public static interface DocumentObserver {
    	void onClose();
    	void onRead();
    }

    
}
