package zdb.core;

import java.io.File;
import java.io.IOException;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.lucene.analysis.Analyzer;
import org.apache.lucene.document.Document;
import org.apache.lucene.document.Field;
import org.apache.lucene.index.IndexWriter;
import org.apache.lucene.index.Term;
import org.apache.lucene.index.IndexReader.FieldOption;
import org.apache.lucene.search.Hit;
import org.apache.lucene.search.Hits;
import org.apache.lucene.search.IndexSearcher;
import org.apache.lucene.search.Query;
import org.apache.lucene.search.Sort;
import org.apache.lucene.search.SortField;
import org.apache.lucene.store.FSDirectory;

import zdb.client.Document.Index;
import zdb.exceptions.UnexpectedException;

public class Indexer {

    private static Map<File, IndexWriter> indexs = new HashMap<File, IndexWriter>();

    private static Map<File, IndexSearcher> searchers = new HashMap<File, IndexSearcher>();

    private static Map<File, Long> writerStats = new HashMap<File, Long>();

    private static long OPTIMIZE_THRESHOLD;

    private static String ANALYSER_CLASS;
    

    static {
        OPTIMIZE_THRESHOLD = Long.parseLong(System.getProperty("zdb.optimize", "1000"));
        ANALYSER_CLASS = System.getProperty("zdb.analyser.class", "org.apache.lucene.analysis.standard.StandardAnalyzer");
    }

    private static Analyzer getAnalyser() {
        try {
            Class clazz = Class.forName(ANALYSER_CLASS);
            return (Analyzer) clazz.newInstance();
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }
    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~ Indexs management
    private static IndexWriter getIndexWriterForBucket(File bucket) {
        try {
            if (!indexs.containsKey(bucket)) {
                synchronized (Indexer.class) {
                    if (!indexs.containsKey(bucket)) {
                        File indexRoot = new File(bucket, "_index");
                        if (!indexRoot.exists()) {
                            indexRoot.mkdirs();
                        }
                        // Set a config here
                        if (new File(indexRoot, "write.lock").exists()) {
                            new File(indexRoot, "write.lock").delete();
                        }
                        IndexWriter writer = new IndexWriter(FSDirectory.getDirectory(indexRoot), true, getAnalyser());
                        indexs.put(bucket, writer);
                    }
                }
            }
            IndexWriter wr = indexs.get(bucket);
            long count = getStat(bucket).longValue();
            if (count > OPTIMIZE_THRESHOLD) {
                long start = System.currentTimeMillis();
                wr.optimize(true);
                writerStats.put(bucket, new Long(0));
                System.out.println("Optimized index " + bucket.getName() + " in " + (System.currentTimeMillis() - start) + " ms");
            }
            return wr;
        } catch (Exception e) {
            throw new UnexpectedException("Cannot open index", e);
        }
    }

    private static Long getStat(File file) {
        if (!writerStats.containsKey(file)) {
            writerStats.put(file, new Long(0));
        }
        return writerStats.get(file);
    }

    private static IndexSearcher getIndexSearcherForBucket(File bucket, boolean forceRefresh) {
        try {
            if (!searchers.containsKey(bucket) || forceRefresh) {
                synchronized (Indexer.class) {
                    IndexSearcher old = searchers.get(bucket);
                    File indexRoot = new File(bucket, "_index");
                    if (indexRoot.exists()) {
                        IndexSearcher indexSearcher = new IndexSearcher(FSDirectory.getDirectory(indexRoot));
                        searchers.put(bucket, indexSearcher);
                    }
                    if (old != null) {
                        old.close();
                    }
                }
            }
            return searchers.get(bucket);
        } catch (Exception e) {
            throw new UnexpectedException("Cannot open index", e);
        }
    }

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~ Lucene document
    public static Index parseIndex(String indexString) {
        int i = indexString.indexOf("=");
        assert i > 0 : "BAD Index key";
        String declaration = indexString.substring(0, i);
        String value = indexString.substring(i + 1);
        String[] parts = declaration.split(";");
        String key = parts[parts.length - 1];
        zdb.client.Document.Index index = new zdb.client.Document.Index(key, value);
        for (String part : parts) {
            if (part.equals("STORED")) {
                index.stored = true;
            }
            if (part.equals("INDEXED")) {
                index.indexed = true;
            }
            if (part.equals("FULLTEXT")) {
                index.fulltext = true;
            }
            if (part.equals("EXACT")) {
                index.exact = true;
            }
        }
        return index;
    }

    public static Document toDocument(String id, List<String> indexes) {
        Document luceneDoc = new Document();
        luceneDoc.add(new Field("_docID", id, Field.Store.YES, Field.Index.UN_TOKENIZED));

        for (String indexInfos : indexes) {
            Index newindex = parseIndex(indexInfos);
            Field.Index luceneIndex = Field.Index.UN_TOKENIZED;
            if (!newindex.indexed) {
                luceneDoc.add(new Field(newindex.name, newindex.value.toString(), Field.Store.NO, Field.Index.NO));
            } else if (newindex.fulltext) {
                luceneDoc.add(new Field(newindex.name, newindex.value.toString(), Field.Store.YES, Field.Index.TOKENIZED));
            } else {
                luceneDoc.add(new Field(newindex.name, newindex.value.toString(), Field.Store.YES, Field.Index.UN_TOKENIZED));
            // Field.Index ==  (Boolean) pi[2] ? ((Boolean) pi[3] ? Field.Index.TOKENIZED : Field.Index.UN_TOKENIZED) : Field.Index.NO;
            }
        }
        return luceneDoc;
    }

    // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~ API
    public static void storeToIndex(File bucket, String id, Document document) throws 
            Exception {
        IndexWriter indexWriter = getIndexWriterForBucket(bucket);
        indexWriter.deleteDocuments(new Term("_docID", id));
        indexWriter.addDocument(document);
        indexWriter.flush();
        long opCount = getStat(bucket).longValue();
        opCount++;
        writerStats.put(bucket, new Long(opCount));
        getIndexSearcherForBucket(bucket, true);
    }

    public static void deleteFromIndex(File bucket, String id) throws Exception {
        IndexWriter indexWriter = getIndexWriterForBucket(bucket);
        indexWriter.deleteDocuments(new Term("_docID", id));
        indexWriter.flush();
        getIndexSearcherForBucket(bucket, true);
    }

    public static void initialize(File bucket) {
        getIndexWriterForBucket(bucket);
    }

    public static Document searchOne(File bucket, Query query) {
        try {
            IndexSearcher searcher = getIndexSearcherForBucket(bucket, false);
            if (searcher != null) {
                Hits hits = searcher.search(query);
                Hit hit = (Hit) hits.iterator().next();
                return hit.getDocument();
            }
            return null;
        } catch (IOException e) {
            throw new UnexpectedException("IO error", e);
        }
    }

    public static Hits search(File bucket, Query query, Sort sort) {
        try {
            IndexSearcher searcher = getIndexSearcherForBucket(bucket, false);
            Collection<String> fields = (Collection<String>) searcher.getIndexReader().getFieldNames(FieldOption.ALL);
            SortField[] sorts = sort.getSort();
            for (SortField sortField : sorts) {
                if (!fields.contains(sort)) {
                    return searcher.search(query, null, new Sort(new String[0]));
                }
            }
            if (searcher != null) {
                return searcher.search(query, null, sort);
            }
            return null;
        } catch (IOException e) {
            throw new UnexpectedException("IO error", e);
        }
    }

    public static void shutdown() throws Exception {
        for (IndexWriter writer : indexs.values()) {
            writer.close();
        }
        for (IndexSearcher searcher : searchers.values()) {
            searcher.close();
        }
    }
}
