package zdb.core;

import java.io.BufferedReader;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReadWriteLock;
import java.util.concurrent.locks.ReentrantReadWriteLock;
import java.util.regex.Pattern;

import org.apache.log4j.Logger;
import org.apache.lucene.analysis.KeywordAnalyzer;
import org.apache.lucene.document.Document;
import org.apache.lucene.document.Field;
import org.apache.lucene.queryParser.ParseException;
import org.apache.lucene.queryParser.QueryParser;
import org.apache.lucene.search.Hits;
import org.apache.lucene.search.Query;
import org.apache.lucene.search.Sort;

import zdb.client.Document.Index;
import zdb.exceptions.BucketDoesNotExistException;
import zdb.exceptions.DocumentAlreadyExistsException;
import zdb.exceptions.DocumentDoesNotExistException;
import zdb.exceptions.LockedException;
import zdb.exceptions.UnexpectedException;
import zdb.exceptions.ZDBException;

public class Store {

    static Logger logger = Logger.getLogger("zdb");

    static File data;

    static boolean started;

    static Pattern validID = Pattern.compile("^[^_][-_%0-9a-zA-Z]*$");

    static Pattern validBucket = Pattern.compile("^[^_][-_%/0-9a-zA-Z]*$");

    public static boolean init(File data) {
        if (!data.exists()) {
            logger.info("Creating datastore at " + data.getAbsolutePath());
            if (!data.mkdirs()) {
                logger.error("Cannot init datastore at " + data.getAbsolutePath());
            }
        }
        Store.data = data;
        started = true;
        return true;
    }

    public static void save(String bucket, String id, InputStream data, String mimeType, List<String> indexes, Boolean update) {
        assert validID.matcher(id).matches();
        File bucketDir = getBucket(bucket);
        String fileName = id;
        File content = new File(bucketDir, fileName + ".content");
        File meta = new File(bucketDir, fileName + ".meta");
        // Meta data
        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        try {
            byte[] breakLine = "\n".getBytes("utf-8");
            baos.write(mimeType.getBytes("utf-8"));
            baos.write(breakLine);
            baos.write(id.getBytes("utf-8"));
            baos.write(breakLine);
            for (String index : indexes) {
                baos.write(index.getBytes("utf-8"));
                baos.write(breakLine);
            }
        } catch (IOException e) {
            // no IO Exception
            }
        byte[] metaData = baos.toByteArray();
        // Lock and write
        Lock lock = getLock(bucket, id).writeLock();
        try {
            if (lock.tryLock(1, TimeUnit.SECONDS)) {
                try {
                    if (update != null) {
                        if (update) {
                            if (!meta.exists()) {
                                throw new DocumentDoesNotExistException(bucket, id);
                            }
                        } else {
                            if (meta.exists()) {
                                throw new DocumentAlreadyExistsException(bucket, id);
                            }
                        }
                    }
                    // Add to index
                    Document document = Indexer.toDocument(id, indexes);
                    Indexer.storeToIndex(bucketDir, id, document);
                    // write content
                    byte[] buffer = new byte[100 * 1024];
                    int read = -1;
                    FileOutputStream fos = new FileOutputStream(content);
                    while ((read = data.read(buffer)) > 0) {
                        fos.write(buffer, 0, read);
                    }
                    fos.close();
                    // write meta
                    fos = new FileOutputStream(meta);
                    fos.write(metaData);
                    fos.close();

                } catch (ZDBException e) {
                    throw e;
                } catch (Exception e) {
                    throw new UnexpectedException("Cannot write", e);
                } finally {
                    lock.unlock();
                }
            } else {
                throw new LockedException();
            }
        } catch (InterruptedException e) {
            throw new UnexpectedException("Thread was interrupted while whaiting write lock", e);
        }
    }

    public static boolean exists(String bucket, String id) {
        File bucketDir = getBucket(bucket);
        File content = new File(bucketDir, id + ".content");
        return content.exists();
    }

    public static ReadResult read(String bucket, String id, boolean metas) {
        File bucketDir = getBucket(bucket);
        String fileName = id;
        File content = new File(bucketDir, fileName + ".content");
        File meta = new File(bucketDir, fileName + ".meta");
        if (!meta.exists() || !content.exists()) {
            throw new DocumentDoesNotExistException(bucket, id);
        }
        // Lock and read
        Lock lock = getLock(bucket, id).readLock();
        try {
            if (lock.tryLock(1, TimeUnit.SECONDS)) {
                try {
                    ReadResult result = new ReadResult();
                    result.id = id;
                    result.lock = lock;
                    // read mimeType
                    if (!metas) {
                        result.mimeType = readMime(meta);
                    } else {
                        readMetas(meta, result);
                    // read content
                    }
                    result.size = content.length();
                    result.file = content;
                    return result;
                } catch (Exception e) {
                    throw new UnexpectedException("Cannot read", e);
                }
            } else {
                throw new LockedException();
            }
        } catch (InterruptedException e) {
            throw new UnexpectedException("Thread was interrupted while waiting write lock", e);
        }
    }

    private static void readMetas(File meta, ReadResult result) throws Exception {
        result.indexes = new ArrayList<Index>();
        BufferedReader rd = new BufferedReader(new FileReader(meta));
        result.mimeType = rd.readLine();
        rd.readLine();//id
        while (true) {
            String tmp = rd.readLine();
            if (tmp == null) {
                break;
            } else {
                String[] kv = tmp.split("=", 2);
                String[] metas = kv[0].split(";");
                boolean indexed = false;
                boolean stored = false;
                boolean fulltext = false;
                boolean exact = false;
                for (int i = 0; i < metas.length - 1; i++) {
                    if (metas[i].equals("INDEXED")) {
                        indexed = true;
                    } else if (metas[i].equals("STORED")) {
                        stored = true;
                    } else if (metas[i].equals("FULLTEXT")) {
                        fulltext = true;
                    } else if (metas[i].equals("EXACT")) {
                        exact = true;
                    }
                }
                Index index = new Index(metas[metas.length - 1], kv[1], stored, indexed, fulltext, exact);
                result.indexes.add(index);
            }
        }
        rd.close();
    }

    private static String readMime(File meta) throws Exception {
        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        FileInputStream fis = new FileInputStream(meta);
        int read = 0;
        while ((read = fis.read()) != -1) {
            if (read == 10) {
                break;
            }
            baos.write(read);
        }
        String result = new String(baos.toByteArray(), "utf-8");
        fis.close();
        return result;
    }

    public static void delete(String bucket, String id) {
        assert validID.matcher(id).matches();
        File bucketDir = getBucket(bucket);
        File content = new File(bucketDir, id + ".content");
        File meta = new File(bucketDir, id + ".meta");
        Lock lock = getLock(bucket, id).writeLock();
        try {
            if (lock.tryLock(1, TimeUnit.SECONDS)) {
                try {
                    if (!meta.exists()) {
                        throw new DocumentDoesNotExistException(bucket, id);
                    }
                    meta.delete();
                    content.delete();
                    Indexer.deleteFromIndex(bucketDir, id);
                } catch (Exception e) {
                    throw new UnexpectedException("Cannot write", e);
                }
            } else {
                throw new LockedException();
            }
        } catch (InterruptedException e) {
            throw new UnexpectedException("Thread was interrupted while waiting write lock", e);
        } finally {
            lock.unlock();
        }
    }

    public static class ReadResult {

        public File file;

        public String id;

        public String mimeType;

        public long size;

        public List<Index> indexes;

        Lock lock;

        public void release() {
            lock.unlock();
        }

        @Override
        public String toString() {
            StringBuilder res = new StringBuilder();
            res.append("ID : " + id);
            res.append("\nMIME : " + mimeType);
            res.append("\nSIZE : " + size);
            return res.toString();
        }
    }

    public static class QueryResult {
        public Document luceneDocument;
        public float score;
        public List<zdb.client.Document.Index> indexes = new ArrayList<zdb.client.Document.Index>();

        public String docID() {
            return luceneDocument.get("_docID");
        }
    }
    
    public static List<QueryResult> search(String bucket, Query query, Sort sort, int offset, int limit, String retrieve) {
        try {
            List<QueryResult> results = new ArrayList<QueryResult>();
            File bucketDir = getBucket(bucket);
            Hits hits = Indexer.search(bucketDir, query, sort);
            if( hits == null )
                return results;
            
            int l = hits.length();
            if (offset > l) {
                return results;
            }

            List<String> indexes = new ArrayList<String>();
            if (retrieve != null) {
                String[] tmp = retrieve.split(",");
                for (int i = 0; i < tmp.length; i++) {
                    indexes.add(tmp[i].trim());
                }
            }
            for (int i = offset; i < (offset + limit > l ? l : offset + limit); i++) {
                QueryResult qresult = new QueryResult();
                qresult.luceneDocument = hits.doc(i);
                qresult.score = hits.score(i);
                for (String index : indexes) {
                    qresult.indexes.add(new zdb.client.Document.Index(index, hits.doc(i).get(index)));
                    //docIndexes.add(new zdb.client.Document.Index(index, document.get(index)));
                }
                //result.put(document.get("_docID"), docIndexes);
                results.add(qresult);
            }
            return results;
        } catch (Exception e) {
            throw new UnexpectedException("Corrupted index", e);
        }
    }

    public static int count(String bucket, Query query) {
        File bucketDir = getBucket(bucket);
        Hits hits = Indexer.search(bucketDir, query, Sort.INDEXORDER);
        return hits == null ? 0 : hits.length();
    }

    public static void createBucket(String name) {
        try {
            getBucket(name);
            return;
        } catch (BucketDoesNotExistException e) {
            assert validBucket.matcher(name).matches();
            File bucketFile = new File(data, name);
            if (bucketFile.mkdirs()) {
                Indexer.initialize(bucketFile);
                buckets.put(name, bucketFile);
                return;
            }
            throw new UnexpectedException("Cannot create bucket", e);
        }
    }

    public static List<zdb.client.Document.Index> fetchAllIndexes(String bucket, String id) {
        List<zdb.client.Document.Index> indexes = new ArrayList<zdb.client.Document.Index>();
        try {
            Query q = new QueryParser("", new KeywordAnalyzer()).parse("_docID:" + id);
            File bucketDir = getBucket(bucket);
            Document doc = Indexer.searchOne(getBucket(bucket), q);
            List<Field> fields = doc.getFields();
            for (Field field : fields) {
                zdb.client.Document.Index index = new zdb.client.Document.Index(field.name(), field.stringValue());
                index.indexed = field.isIndexed();
                index.stored = true;
                indexes.add(index);
            }
        } catch (ParseException e) {
            throw new RuntimeException(e);
        }
        return indexes;
    }

    static File getBucket(String bucket) {
        if (!started) {
            throw new UnexpectedException("ZDB not started");
        }
        /*if (!bucket.matches("^[^/_][-_/@a-zA-Z0-9]+[^/]$")) {
        throw new InvalidNameException("Invalid bucket name : " + bucket);
        }*/
        if (buckets.containsKey(bucket)) {
            return buckets.get(bucket);
        }
        File bucketDir = new File(data, bucket);
        if (bucketDir.exists() && bucketDir.isDirectory()) {
            buckets.put(bucket, bucketDir);
            return bucketDir;
        }
        throw new BucketDoesNotExistException(bucket);
    }

    static Map<String, File> buckets = new HashMap<String, File>();

    static ReadWriteLock getLock(String bucket, String id) {
        String lockId = bucket + "/" + id;
        if (locks.containsKey(lockId)) {
            return locks.get(lockId);
        }
        synchronized (Store.class) {
            if (locks.containsKey(lockId)) {
                return locks.get(lockId);
            }
            ReadWriteLock lock = new ReentrantReadWriteLock();
            locks.put(lockId, lock);
            return lock;
        }
    }

    static Map<String, ReadWriteLock> locks = new HashMap<String, ReadWriteLock>();
}
